from tqdm import tqdm
import json
import numpy as np
import pandas as pd
from molbar.barcode import get_molbars_from_coordinates
from rdkit import Chem
import matplotlib.pyplot as plt
from rdkit.Chem  import rdDetermineBonds
from rdkit import RDLogger
import subprocess
import os
import shutil   
from molbar.helper.printer import Printer
from joblib import Parallel, delayed
RDLogger.DisableLog('rdApp.*') 

def obabel_interface(path):
     
    proc = subprocess.Popen(f"obabel {path} -o inchi", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    stdout, stderr = proc.communicate(timeout=60)
    
    stdout_str = stdout.decode("utf-8")

    identifier = stdout_str.split("\t")[0]

    return identifier

def get_inchi_from_obabel(list_of_coordinates, list_of_elements, threads=8):

    tmp_dir = "./tmp/"

    if os.path.exists(tmp_dir):

        shutil.rmtree(tmp_dir)

    os.mkdir(tmp_dir)

    for idx, coordinates, elements in zip(range(len(list_of_elements)), list_of_coordinates, list_of_elements):

        n_atoms = len(elements)

        path = tmp_dir + f"{idx}.xyz"

        Printer(n_atoms=n_atoms, energy="", coordinates=coordinates, elements=elements, path=path).print()

    all_files = [os.path.join(tmp_dir, f"{x}.xyz") for x in range(len(list_of_elements))]

    results = Parallel(n_jobs=threads, prefer="threads")(delayed(obabel_interface)(file)
                                     for file in tqdm(all_files))

    return results

def get_inchi_for_permutations_obabel(df, n_permutations=5):

    df["InChI_from_obabel"] = get_inchi_from_obabel(df["coordinates"], df["elements"], threads=8)

    for iteration  in range(1, n_permutations+1):

        print(f"for permutation {iteration}")

        df[f"InChI_from_obabel_{iteration}"] = get_inchi_from_obabel(df[f"coordinates_permutation_{iteration}"], df[f"elements_permutation_{iteration}"], threads=8)

    return df


def write_xyz_file(elements, coordinates):
    if len(elements) != len(coordinates):
        raise ValueError("Number of elements and coordinates must be the same")

    output = ""

    num_atoms = len(elements)
    output += str(num_atoms) + "\n \n"

    for i in range(num_atoms):
        element = elements[i]
        x, y, z = coordinates[i]
        output += f"{element} {x} {y} {z}\n"

    return output

def permutate_df(df, n_permutations=5):

    df["n_atoms"] = df.apply(lambda x: len(x["coordinates"]), axis=1)

    for iteration  in tqdm(range(1, n_permutations+1)):
            
        # Create a permutation key
        df[f"permutation_key_{iteration}"] = df["n_atoms"].apply(np.random.permutation)

        # Apply permutations directly using NumPy
        df[f"coordinates_permutation_{iteration}"] = df.apply(lambda x: np.array(x["coordinates"])[x[f"permutation_key_{iteration}"]], axis=1)
        
        df[f"elements_permutation_{iteration}"] = df.apply(lambda x: np.array(x["elements"])[x[f"permutation_key_{iteration}"]], axis=1)

    return df

def sample_df(df, sample_size):

    return df.sample(sample_size)

def get_inchi_from_rdkit(coordinates, elements):

    xyzblock = write_xyz_file(elements, coordinates)

    raw_mol = Chem.MolFromXYZBlock(xyzblock)

    conn_mol = Chem.Mol(raw_mol)

    try:
        
        rdDetermineBonds.DetermineConnectivity(conn_mol)

        mol = Chem.RemoveAllHs(conn_mol)

        inchi = Chem.MolToInchi(mol)

        return inchi

    except:

        return ""
        
def get_inchi_for_permutations_rdkit(df, n_permutations=5):

    df["InChI_from_rdkit"] = df.progress_apply(lambda x: get_inchi_from_rdkit(x["coordinates"], x["elements"]), axis=1)

    for iteration  in range(1, n_permutations+1):

        print(f"for permutation {iteration}")

        df[f"InChI_from_rdkit_{iteration}"] = df.progress_apply(lambda x: get_inchi_from_rdkit(x[f"coordinates_permutation_{iteration}"], x[f"elements_permutation_{iteration}"]), axis=1)

    return df

def get_molbars_for_permutations(df, n_permutations=5, threads=8):

    df["MolBar"] = get_molbars_from_coordinates(df["coordinates"].values, df["elements"].values, progress=True, threads=threads)

    for iteration in range(1, n_permutations+1):

        print(f"for permutation {iteration}")

        df[f"molbars_permutation_{iteration}"] = get_molbars_from_coordinates(df[f"coordinates_permutation_{iteration}"].values, df[f"elements_permutation_{iteration}"].values, progress=True, threads=threads)

    return df

def analyze_permutation_failures(df, n_permutations=5):

    inchi_rdkit_perm_failure = []

    molbar_perm_failure = []

    inchi_obabel_perm_failure = []

    for iteration in tqdm(range(1, n_permutations+1)):

        df[f"InChI_from_rdkit_perm_fail_{iteration}"] = 0

        df.loc[df["InChI_from_rdkit"] != df[f"InChI_from_rdkit_{iteration}"], f"InChI_from_rdkit_perm_fail_{iteration}"] = 1

        if iteration == 1:

            inchi_rdkit_perm_failure.append(df[f"InChI_from_rdkit_perm_fail_{iteration}"].sum())

        else:

            prev_fail = inchi_rdkit_perm_failure[-1]

            inchi_rdkit_perm_failure.append(df[f"InChI_from_rdkit_perm_fail_{iteration}"].sum()+prev_fail)

        df[f"molbars_perm_fail_{iteration}"] = 0

        df.loc[df["MolBar"] != df[f"molbars_permutation_{iteration}"], f"molbars_perm_fail_{iteration}"] = 1

        if iteration == 1:

            molbar_perm_failure.append(df[f"molbars_perm_fail_{iteration}"].sum())

        else:

            prev_fail = molbar_perm_failure[-1]

            molbar_perm_failure.append(df[f"molbars_perm_fail_{iteration}"].sum()+prev_fail)

        df[f"InChI_from_obabel_perm_fail_{iteration}"] = 0

        df.loc[df["InChI_from_obabel"] != df[f"InChI_from_obabel_{iteration}"], f"InChI_from_obabel_perm_fail_{iteration}"] = 1

        if iteration == 1:

            inchi_obabel_perm_failure.append(df[f"InChI_from_obabel_perm_fail_{iteration}"].sum())

        else:

            prev_fail = inchi_obabel_perm_failure[-1]

            inchi_obabel_perm_failure.append(df[f"InChI_from_obabel_perm_fail_{iteration}"].sum()+prev_fail)

    print(f"Summed MolBar failures: {molbar_perm_failure}")
    print(f"Summed InChI failures from RDKit: {inchi_rdkit_perm_failure}")
    print(f"Summed InChI failures from Open Babel: {inchi_obabel_perm_failure}")

    return inchi_rdkit_perm_failure, molbar_perm_failure, inchi_obabel_perm_failure


with open("element_data.json") as json_file:

    data = json.load(json_file)

def nuclear_rep_(coordinates, elements):

    nuclear_charges = [data["atomic_numbers"][element] for element in elements]

    nuclear_repulsion = 0
    for i in range(len(elements)):
        for j in range(i+1, len(elements)):
            dist = np.linalg.norm(coordinates[i] - coordinates[j])
            nuclear_repulsion += nuclear_charges[i] * nuclear_charges[j] / dist

    return nuclear_repulsion

def get_permutation_nuclear_rep(df, n_permutations=5):

    V_nuc = []

    for iteration in tqdm(range(1, n_permutations+1)):

        df[f"nuclear_rep_permutation_{iteration}"] = df.apply(lambda x: nuclear_rep_(x[f"coordinates_permutation_{iteration}"], x[f"elements_permutation_{iteration}"]), axis=1)

        V_nuc.append(df[f"nuclear_rep_permutation_{iteration}"].mean())

    return V_nuc

def get_permutation_plot(inchi_rdkit_perm_failure, molbar_perm_failure, inchi_obabel_perm_failure, V_nuc):

    plt.style.use("rwth_colors.mplstyle")

    fig, ax1 = plt.subplots(figsize=(10, 8))
    ax2 = ax1.twinx()
    x = list(range(1, 6))
    ax1.plot(x, inchi_obabel_perm_failure, "o", label="InChI with Open Babel from .xyz", ms=12, color="#CC071E")
    ax1.plot(x, inchi_rdkit_perm_failure, "o", label="InChI with RDKit from .xyz", ms=12, color="#F6A800")
    ax1.plot(x, molbar_perm_failure, "*", label="MolBar from .xyz", ms=12, color="#00549F")
    ax1.set_xlabel("Permutations", fontsize=18)
    ax1.set_ylabel("Summed Permutation Failures", fontsize=18)
    ax1.tick_params(axis="both", labelsize=18)
    ax1.legend(fontsize=18)
    ax2.plot(x, V_nuc, "k:", ms=12)
    ax2.set_ylabel(r"Mean Nuclear Repulsion Energy [$E_h$]", fontsize=18)
    ax2.tick_params(axis="y", labelsize=18)
    #ax2.legend(fontsize=18)

    plt.xticks(x, fontsize=18)

    plt.savefig("../figures/permutation_test.png", bbox_inches="tight")




if __name__ == "__main__":

    tqdm.pandas()

    df = pd.read_pickle("./data/combined_mols.pkl")

    df = sample_df(df, 10)

    print("Permutating")

    df = permutate_df(df)

    print("Getting MolBars")

    df = get_molbars_for_permutations(df)

    print("Getting InChIs from RDKit")

    df = get_inchi_for_permutations_rdkit(df)

    print("Getting InChIs from Open Babel")

    df = get_inchi_for_permutations_obabel(df)

    df.to_pickle("permutation_data_test.pkl")

    print("Analyzing permutation failures")

    inchi_rdkit_perm_failure, molbar_perm_failure, inchi_obabel_perm_failure = analyze_permutation_failures(df)

    print("Getting nuclear repulsion")

    V_nuc = get_permutation_nuclear_rep(df)

    print("Plotting")

    get_permutation_plot(inchi_rdkit_perm_failure, molbar_perm_failure,inchi_obabel_perm_failure, V_nuc)





